use std::{
    fs::File,
    io::{prelude::*, LineWriter},
    path::Path,
};

pub type Result<T, E = Error> = std::result::Result<T, E>;
#[derive(Debug, thiserror::Error)]
pub enum Error {
    #[error("IoError")]
    IoError { source: std::io::Error },
}

pub fn create(crate_name: &str, feature_mod_names: &[(String, String)], path: &Path) -> Result<()> {
    let file = File::create(path).map_err(|source| Error::IoError { source })?;
    let mut file = LineWriter::new(file);
    let default = get_default_feature(feature_mod_names);
    file.write_all(
        format!(
            r#"# generated by AutoRust
[package]
name = "{}"
version = "0.1.0"
edition = "2018"
license = "MIT"
description = "generated REST API bindings"

[dependencies]
azure_core = {{ path = "../../../sdk/core", version = "0.1", default-features = false }}
serde = {{ version = "1.0", features = ["derive"] }}
serde_json = "1.0"
bytes = "1.0"
thiserror = "1.0"
http = "0.2"
url = "2.2"
futures = "0.3"

[dev-dependencies]
azure_identity = {{ path = "../../../sdk/identity", version = "0.1" }}
tokio = {{ version = "1.0", features = ["macros"] }}

[package.metadata.docs.rs]
all-features = true

[features]
default = ["{}", "enable_reqwest"]
enable_reqwest = ["azure_core/enable_reqwest"]
enable_reqwest_rustls = ["azure_core/enable_reqwest_rustls"]
no-default-version = []
"#,
            crate_name, default
        )
        .as_bytes(),
    )
    .map_err(|source| Error::IoError { source })?;

    for (feature_name, _mod_name) in feature_mod_names {
        file.write_all(format!("\"{}\" = []\n", feature_name).as_bytes())
            .map_err(|source| Error::IoError { source })?;
    }
    Ok(())
}

fn get_default_feature(feature_mod_names: &[(String, String)]) -> String {
    let default = feature_mod_names
        .iter()
        .map(|(feature, _)| feature)
        .find(|feature| !feature.contains("preview"));
    match default {
        Some(default) => default.clone(),
        None => feature_mod_names[0].0.clone(),
    }
}
